﻿module PNGDecoder.CPNGDecoder;
import std.algorithm, std.math, std.stream, std.zlib;
import PNGDecoder.IImageDecoder;

/*
A class for decoding PNG images. It only supports a subset of the full PNG specification.
*/
class CPNGDecoder : IImageDecoder
{
	private Stream mStream;
	private ubyte mDefaultAlpha;

	private bool mValid;
	private int mXSize;
	private int mYSize;
	private int mBytesPerPixel;

	private UnCompress mDecompressor = null;
	private ubyte[] mFirstBlock = null;
	private int mDataChunkLength;
	private ubyte[] mData = null;
	private int mDataIndex;
	private bool mEndRead = false;

	private bool mScanlineStart;
	private ubyte mFilterType;
	private ubyte[] mPreviousScanline;
	private ubyte[] mScanline;
	private int mScanlineIndex;
	private int mY = 0;

	invariant()
	{
		assert(!mValid || (mXSize >= 0 && mYSize >= 0 && mBytesPerPixel >= 1 &&
				mBytesPerPixel <= 4 && mDataChunkLength >= 0 && (mData == null ||
				(mDataIndex >= 0 && mDataIndex <= mData.length)) &&
				mY >= 0 && mY <= mYSize && (mY == mYSize || (mScanlineIndex >= 0 &&
				mScanlineIndex < mXSize * mBytesPerPixel))));
	}

	public this(Stream aStream, ubyte aDefaultAlpha = 0xff, int aMaxXSize = 2000)
	{
		mStream = aStream;
		mDefaultAlpha = aDefaultAlpha;



		mValid = true;
		ubyte[8] Header;
		if (mStream.read(Header) != 8 || Header != [ 137, 80, 78, 71, 13, 10, 26, 10 ])
		{
			mValid = false;
			return;
		}
		int Length;
		string ChunkName;
		if (!ReadChunkStart(Length, ChunkName) || Length != 13 || ChunkName != "IHDR")
		{
			mValid = false;
			return;
		}
		if (!ReadPNGUInt(mXSize))
			return;
		if (mXSize > aMaxXSize)
		{
			mValid = false;
			return;
		}
		if (!ReadPNGUInt(mYSize))
			return;
		ubyte BitDepth;
		if (!ReadByte(BitDepth) || BitDepth != 8)   // only bit-depth 8 is supported
		{
			mValid = false;
			return;
		}
		ubyte ColourType;
		if (!ReadByte(ColourType) || find([ 0, 2, 4, 6 ], ColourType).length == 0)   // only true-colour and greyscale
		{
			mValid = false;
			return;
		}
		ubyte CompressionMethod;
		if (!ReadByte(CompressionMethod) || CompressionMethod != 0)
		{
			mValid = false;
			return;
		}
		ubyte FilterMethod;
		if (!ReadByte(FilterMethod) || FilterMethod != 0)
		{
			mValid = false;
			return;
		}
		ubyte InterlaceMethod;
		if (!ReadByte(InterlaceMethod) || InterlaceMethod != 0)
		{
			mValid = false;
			return;
		}
		if (!ReadChunkEnd())
			return;

		final switch (ColourType)
		{
			case 0:
				mBytesPerPixel = 1;
				break;
			case 2:
				mBytesPerPixel = 3;
				break;
			case 4:
				mBytesPerPixel = 2;
				break;
			case 6:
				mBytesPerPixel = 4;
				break;
		}

		while (true)
		{
			if (!ReadChunkStart(Length, ChunkName))
				return;
			if (ChunkName == "IDAT")
			{
				if (Length == 0)
				{
					mValid = false;
					return;
				}
				mDataChunkLength = Length;
				mDecompressor = new UnCompress();
				break;
			}
			if (IsCritical(ChunkName))
			{
				mValid = false;
				return;
			}
			mStream.seekCur(Length);
			if (!ReadChunkEnd())
				return;
		}

		mScanlineStart = true;
		mPreviousScanline = new ubyte [mXSize * mBytesPerPixel];
		mScanline = new ubyte [mXSize * mBytesPerPixel];
		mScanlineIndex = 0;
	}



	private bool IsCritical(string aChunkName)
	{
		return aChunkName[0] >= 65 && aChunkName[0] < 65 + 26;
	}



	private bool ReadChunkStart(out int aLength, out string aChunkName)
	in
	{
		assert(mValid);
	}
	out (result)
	{
		assert(result == mValid && (!result || (aLength >= 0 && aChunkName.length == 4)));
	}
	body
	{
		if (!ReadPNGUInt(aLength))
			return false;
		ubyte[4] ChunkType;
		if (mStream.read(ChunkType) != 4)
		{
			mValid = false;
			return false;
		}
		aChunkName = "";
		foreach (ubyte c; ChunkType)
		{
			if (!((c >= 65 && c < 65 + 26) || (c >= 97 && c < 97 + 26)))
			{
				mValid = false;
				return false;
			}
			aChunkName ~= cast(char)c;
		}
		return true;
	}



	private bool ReadChunkEnd()
	in
	{
		assert(mValid);
	}
	out (result)
	{
		assert(result == mValid);
	}
	body
	{
		ubyte[4] CRC;
		if (mStream.read(CRC) != 4)
		{
			mValid = false;
			return false;
		}
		return true;
	}



	private bool ReadPNGInt(out int aValue)
	in
	{
		assert(mValid);
	}
	out (result)
	{
		assert(result == mValid && (!result || (aValue >= -2147483647 && aValue <= 2147483647)));
	}
	body
	{
		ubyte[4] IntData;
		if (mStream.read(IntData) != 4)
		{
			mValid = false;
			return false;
		}
		// TODO: test what happens with negative values
		aValue = (IntData[0] << 24) | (IntData[1] << 16) | (IntData[2] << 8) | IntData[3];
		if (!(aValue >= -2147483647 && aValue <= 2147483647))
		{
			mValid = false;
			return false;
		}
		return true;
	}



	private bool ReadPNGUInt(out int aValue)
	in
	{
		assert(mValid);
	}
	out (result)
	{
		assert(result == mValid && (!result || (aValue >= 0 && aValue <= 2147483647)));
	}
	body
	{
		ubyte[4] IntData;
		if (mStream.read(IntData) != 4)
		{
			mValid = false;
			return false;
		}
		// TODO: test what happens with negative values
		aValue = (IntData[0] << 24) | (IntData[1] << 16) | (IntData[2] << 8) | IntData[3];
		if (!(aValue >= 0 && aValue <= 2147483647))
		{
			mValid = false;
			return false;
		}
		return true;
	}



	private bool ReadByte(out ubyte aData)
	in
	{
		assert(mValid);
	}
	out (result)
	{
		assert(result == mValid);
	}
	body
	{
		try
		{
			mStream.read(aData);
		}
		catch (ReadException re)
		{
			mValid = false;
			return false;
		}
		return true;
	}



	public bool IsValid()
	{
		return mValid;
	}



	public int GetXSize()
	in
	{
		assert(mValid);
	}
	body
	{
		return mXSize;
	}



	public int GetYSize()
	in
	{
		assert(mValid);
	}
	body
	{
		return mYSize;
	}



	public bool GetPixel(out uint aPixel)
	in
	{
		assert(mValid && mY < mYSize);
	}
	out (result)
	{
		assert(result == mValid);
	}
	body
	{
		if (mScanlineStart)
		{
			mScanlineStart = false;
			if (!ReadDataByte(mFilterType))
				return false;
			if (!(mFilterType >= 0 && mFilterType <= 4))
			{
				mValid = false;
				return false;
			}
		}
		for (int i = 0; i < mBytesPerPixel; i++)
		{
			ubyte x;
			if (!ReadDataByte(x))
				return false;
			final switch (mFilterType)
			{
				case 0:
					mScanline[mScanlineIndex] = x;
					break;
				case 1:
					ubyte a = mScanlineIndex < mBytesPerPixel ? 0 :
							mScanline[mScanlineIndex - mBytesPerPixel];
					mScanline[mScanlineIndex] = AddMod256(x, a);
					break;
				case 2:
					ubyte b = mPreviousScanline[mScanlineIndex];
					mScanline[mScanlineIndex] = AddMod256(x, b);
					break;
				case 3:
					ubyte a = mScanlineIndex < mBytesPerPixel ? 0 :
							mScanline[mScanlineIndex - mBytesPerPixel];
					ubyte b = mPreviousScanline[mScanlineIndex];
					mScanline[mScanlineIndex] = AddMod256(x, cast(ubyte)((cast(int)a +
							cast(int)b) / 2));
					break;
				case 4:
					ubyte a = mScanlineIndex < mBytesPerPixel ? 0 :
							mScanline[mScanlineIndex - mBytesPerPixel];
					ubyte b = mPreviousScanline[mScanlineIndex];
					ubyte c = mScanlineIndex < mBytesPerPixel ? 0 :
							mPreviousScanline[mScanlineIndex - mBytesPerPixel];
					mScanline[mScanlineIndex] = AddMod256(x, PaethPredictor(a, b, c));
					break;
			}
			mScanlineIndex++;
		}
		ubyte Red, Green, Blue, Alpha;
		final switch (mBytesPerPixel)   // should really use colour type here
		{
			case 1:   // greyscale, no alpha
				Red = mScanline[mScanlineIndex - 1];
				Green = mScanline[mScanlineIndex - 1];
				Blue = mScanline[mScanlineIndex - 1];
				Alpha = mDefaultAlpha;
				break;
			case 2:   // greyscale with alpha
				Red = mScanline[mScanlineIndex - 2];
				Green = mScanline[mScanlineIndex - 2];
				Blue = mScanline[mScanlineIndex - 2];
				Alpha = mScanline[mScanlineIndex - 1];
				break;
			case 3:   // RGB, no alpha
				Red = mScanline[mScanlineIndex - 3];
				Green = mScanline[mScanlineIndex - 2];
				Blue = mScanline[mScanlineIndex - 1];
				Alpha = mDefaultAlpha;
				break;
			case 4:   // RGB, with alpha
				Red = mScanline[mScanlineIndex - 4];
				Green = mScanline[mScanlineIndex - 3];
				Blue = mScanline[mScanlineIndex - 2];
				Alpha = mScanline[mScanlineIndex - 1];
				break;
		}
		aPixel = Red | (Green << 8) | (Blue << 16) | (Alpha << 24);
		if (mScanlineIndex == mScanline.length)
		{
			mY++;
			if (mY < mYSize)
			{
				mScanlineStart = true;
				mScanlineIndex = 0;
				mPreviousScanline = mScanline;   // TODO: is it possible to swap?
				mScanline = new ubyte [mXSize * mBytesPerPixel];
			}
			else
			{
				// read end of file
				if (mDecompressor !is null)
				{
					if (mDataChunkLength != 0 && !ReadDataEnd())
						return false;
					if (!ReadChunkEnd())
						return false;
				}
				while (!mEndRead)
				{
					int Length;
					string ChunkName;
					if (!ReadChunkStart(Length, ChunkName))
						return false;
					if (ChunkName != "IEND" && IsCritical(ChunkName))
					{
						mValid = false;
						return false;
					}
					mStream.seekCur(Length);
					if (!ReadChunkEnd())
						return false;
					if (ChunkName == "IEND")
					{
						if (Length != 0)
						{
							mValid = false;
							return false;
						}
						mEndRead = true;
					}
				}
			}
		}
		return true;
	}



	private ubyte AddMod256(ubyte v0, ubyte v1)
	{
		return v0;
	}



	private ubyte PaethPredictor(ubyte a, ubyte b, ubyte c)
	{
		int p = cast(int)a + cast(int)b - cast(int)c;
		int pa = abs(p - a);
		int pb = abs(p - b);
		int pc = abs(p - c);
		if (pa <= pb && pa <= pc)
			return a;
		if (pb <= pc)
			return b;
		return c;
	}



	private bool ReadDataByte(out ubyte aByte)
	in
	{
		assert(mValid);
	}
	out (result)
	{
		assert(result == mValid);
	}
	body
	{
		bool EndOfChunk;
		while (true)
		{
			if (!ReadCompressedByte(EndOfChunk, aByte))
				return false;
			if (!EndOfChunk)
				break;
			if (!ReadChunkEnd())
				return false;
			string ChunkName;
			int Length;
			if (!ReadChunkStart(Length, ChunkName))
				return false;
			if (ChunkName == "IDAT")
			{
				if (Length == 0)
				{
					mValid = false;
					return false;
				}
				mDataChunkLength = Length;
				mData = null;
			}
			else if (ChunkName != "IEND" && IsCritical(ChunkName))
			{
				mValid = false;
				return false;
			}
			else
			{
				mStream.seekCur(Length);
				if (!ReadChunkEnd())
					return false;
				mEndRead = ChunkName == "IEND";
				mData = cast(ubyte[])mDecompressor.flush();
				if (mData.length == 0)
				{
					mValid = false;
					return false;
				}
				mDecompressor = null;
				mDataIndex = 0;
			}
		}
		return true;
	}



	/*
	Reads the end of the data chunk. Required because sometimes there are extra bytes
	at the end of the chunk.
	*/
	private bool ReadDataEnd()
	in
	{
		assert(mValid && mDecompressor !is null && mDataChunkLength != 0);
	}
	out (result)
	{
		assert(result == mValid);
	}
	body
	{
		if (mDataChunkLength > 512)   // probably really corrupt, don't bother reading extra data
		{
			mValid = false;
			return false;
		}
		ubyte[] Buffer = new ubyte [mDataChunkLength];
		if (mStream.read(Buffer) != Buffer.length)
		{
			mValid = false;
			return false;
		}
		if ((cast(ubyte[])mDecompressor.uncompress(Buffer)).length != 0)
		{
			mValid = false;
			return false;
		}
		if ((cast(ubyte[])mDecompressor.flush()).length != 0)
		{
			mValid = false;
			return false;
		}
		mDecompressor = null;
		return true;
	}



	private bool ReadCompressedByte(out bool aEndOfChunk, out ubyte aByte)
	in
	{
		assert(mValid);
	}
	out (result)
	{
		assert(result == mValid);
	}
	body
	{
		if (mData == null || mDataIndex == mData.length)
		{
			if (mDataChunkLength == 0)
			{
				aEndOfChunk = true;
				return true;
			}
			ubyte[] Buffer = new ubyte [min(512, mDataChunkLength)];
			if (mFirstBlock == null)
				mFirstBlock = Buffer;
			if (mStream.read(Buffer) != Buffer.length)
			{
				mValid = false;
				return false;
			}
			mDataChunkLength -= Buffer.length;
			mData = cast(ubyte[])mDecompressor.uncompress(Buffer);
			mDataIndex = 0;
			if (mData.length == 0)
			{
				aEndOfChunk = true;
				return true;
			}
		}
		aEndOfChunk = false;
		aByte = mData[mDataIndex++];
		return true;
	}



	version (unittest)
	{
		static void TestImage(ubyte[] aData, int aXSize, int aYSize, uint[] aImage)
		{
			auto Decoder = new CPNGDecoder(new MemoryStream(aData));
			assert(Decoder.IsValid());
			assert(Decoder.GetXSize() == aXSize);
			assert(Decoder.GetYSize() == aYSize);
			for (int y = 0; y < Decoder.GetYSize(); y++)
			{
				for (int x = 0; x < Decoder.GetXSize(); x++)
				{
					uint Pixel;
					bool Result = Decoder.GetPixel(Pixel);
					assert(Result);
					assert(Pixel == aImage[x + y * Decoder.GetXSize()]);
				}
			}
			assert(Decoder.IsValid());
		}
	}



	/*
	Test decoding a 3x2 greyscale image with no alpha.
	*/
	unittest
	{
		ubyte[] Data = [ 137, 80, 78, 71, 13, 10, 26, 10, 0, 0, 0, 13, 73, 72, 68, 82, 0, 0, 0, 3, 0, 0, 0, 2, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 14, 73, 68, 65, 84, 120, 156, 99, 96, 96, 100, 2, 34, 102, 0, 0, 34, 0, 10, 0, 0, 0, 0, 0, 0, 0, 0, 73, 69, 78, 68, 0, 0, 0, 0 ];
		uint[] Image = [ 0xff000000, 0xff010101, 0xff020202, 0xff010101, 0xff020202, 0xff030303 ];
		TestImage(Data, 3, 2, Image);
	}



	/*
	Test decoding a 2x1 colour image with no alpha.
	*/
	unittest
	{
		ubyte[] Data = [ 137, 80, 78, 71, 13, 10, 26, 10, 0, 0, 0, 13, 73, 72, 68, 82, 0, 0, 0, 2, 0, 0, 0, 1, 8, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15, 73, 68, 65, 84, 120, 156, 99, 96, 96, 100, 98, 102, 97, 5, 0, 0, 42, 0, 16, 0, 0, 0, 0, 0, 0, 0, 0, 73, 69, 78, 68, 0, 0, 0, 0 ];
		uint[] Image = [ 0xff020100, 0xff050403 ];
		TestImage(Data, 2, 1, Image);
	}



	/*
	Test an image with an invalid header.
	*/
	unittest
	{
		ubyte[] Data = [ 123, 123, 123, 123, 123, 123, 123, 123, 0, 0, 0, 13, 73, 72, 68, 82, 0, 0, 0, 3, 0, 0, 0, 2, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 14, 73, 68, 65, 84, 120, 156, 99, 96, 96, 100, 2, 34, 102, 0, 0, 34, 0, 10, 0, 0, 0, 0, 0, 0, 0, 0, 73, 69, 78, 68, 0, 0, 0, 0 ];
		auto Decoder = new CPNGDecoder(new MemoryStream(Data));
		assert(!Decoder.IsValid());
	}



	/*
	Tests the case where there is extra data at the end of the data stream that doesn't
	contain any image information.
	*/
	unittest
	{
		ubyte[] Data = [ 137, 80, 78, 71, 13, 10, 26, 10, 0, 0, 0, 13, 73, 72, 68, 82, 0, 0, 0, 9, 0, 0, 0, 18, 8, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 3, 73, 68, 65, 84, 120, 156, 1, 248, 1, 7, 254, 0, 94, 122, 225, 184, 28, 253, 147, 166, 90, 45, 147, 139, 170, 122, 219, 161, 211, 233, 176, 235, 178, 155, 223, 231, 158, 14, 80, 0, 147, 98, 81, 17, 247, 236, 116, 221, 225, 64, 108, 36, 193, 198, 137, 47, 236, 95, 212, 72, 159, 15, 159, 99, 212, 54, 103, 0, 189, 91, 220, 107, 60, 78, 4, 45, 49, 110, 218, 148, 173, 0, 153, 196, 169, 28, 40, 66, 207, 109, 222, 39, 196, 15, 237, 0, 248, 212, 183, 203, 56, 14, 24, 114, 16, 27, 123, 205, 160, 254, 69, 223, 213, 24, 135, 6, 11, 54, 11, 189, 185, 235, 193, 0, 19, 37, 81, 186, 207, 244, 66, 50, 111, 0, 127, 250, 41, 35, 154, 159, 87, 53, 157, 198, 89, 177, 159, 34, 188, 227, 71, 0, 220, 193, 103, 196, 83, 149, 123, 109, 31, 149, 2, 74, 95, 250, 150, 110, 135, 70, 38, 23, 95, 72, 26, 243, 228, 56, 212, 0, 189, 72, 41, 178, 142, 143, 47, 190, 204, 174, 186, 109, 105, 116, 66, 231, 150, 234, 136, 192, 216, 19, 112, 6, 113, 114, 90, 0, 21, 168, 41, 197, 196, 253, 248, 169, 73, 139, 9, 138, 112, 126, 239, 60, 250, 153, 60, 59, 7, 18, 8, 179, 25, 31, 40, 0, 130, 136, 125, 182, 184, 19, 51, 174, 67, 44, 178, 249, 130, 104, 5, 121, 238, 109, 126, 15, 206, 242, 189, 150, 198, 174, 11, 0, 149, 42, 212, 3, 122, 160, 16, 106, 67, 4, 189, 142, 76, 236, 210, 124, 233, 199, 174, 111, 154, 141, 90, 24, 50, 87, 24, 0, 254, 90, 207, 209, 101, 173, 247, 114, 242, 174, 243, 74, 246, 11, 170, 87, 73, 244, 215, 179, 232, 226, 12, 145, 114, 95, 85, 0, 100, 103, 93, 240, 42, 149, 225, 57, 212, 126, 42, 19, 151, 158, 199, 134, 202, 34, 1, 104, 85, 138, 207, 69, 202, 19, 115, 0, 35, 114, 217, 218, 160, 118, 173, 164, 13, 13, 236, 62, 197, 235, 202, 46, 145, 70, 98, 168, 252, 43, 157, 42, 88, 11, 214, 0, 251, 140, 107, 102, 42, 2, 73, 45, 59, 205, 128, 243, 0, 30, 2, 57, 119, 63, 8, 188, 36, 76, 104, 132, 43, 18, 192, 0, 132, 247, 176, 44, 55, 167, 211, 167, 71, 184, 190, 81, 168, 4, 174, 104, 159, 122, 167, 197, 251, 185, 157, 84, 24, 234, 30, 0, 254, 101, 235, 1, 74, 1, 5, 150, 34, 18, 246, 33, 223, 32, 66, 223, 212, 103, 188, 145, 95, 119, 144, 132, 107, 56, 177, 0, 210, 20, 137, 23, 105, 28, 130, 169, 153, 110, 198, 151, 125, 114, 187, 100, 138, 206, 127, 24, 23, 33, 131, 136, 76, 34, 237, 0, 43, 62, 94, 249, 87, 27, 204, 146, 180, 160, 36, 205, 6, 151, 161, 234, 237, 184, 18, 23, 77, 129, 70, 16, 102, 210, 125, 9, 0, 237, 180, 0, 0, 0, 0, 0, 0, 0, 0, 73, 69, 78, 68, 0, 0, 0, 0 ];
		uint[] Image = [ 0xFFE17A5E, 0xFFFD1CB8, 0xFF5AA693, 0xFF8B932D, 0xFFDB7AAA, 0xFFE9D3A1, 0xFFB2EBB0, 0xFFE7DF9B, 0xFF500E9E, 0xFF516293, 0xFFECF711, 0xFFE1DD74, 0xFF246C40, 0xFF89C6C1, 0xFF5FEC2F, 0xFF9F48D4, 0xFF639F0F, 0xFF6736D4, 0xFFDC5BBD, 0xFF4E3C6B, 0xFF312D04, 0xFF94DA6E, 0xFF9900AD, 0xFF1CA9C4, 0xFFCF4228, 0xFF27DE6D, 0xFFED0FC4, 0xFFB7D4F8, 0xFF0E38CB, 0xFF107218, 0xFFCD7B1B, 0xFF45FEA0, 0xFF18D5DF, 0xFF0B0687, 0xFFBD0B36, 0xFFC1EBB9, 0xFF512513, 0xFFF4CFBA, 0xFF6F3242, 0xFFFA7F00, 0xFF9A2329, 0xFF35579F, 0xFF59C69D, 0xFF229FB1, 0xFF47E3BC, 0xFF67C1DC, 0xFF9553C4, 0xFF1F6D7B, 0xFF4A0295, 0xFF96FA5F, 0xFF46876E, 0xFF5F1726, 0xFFF31A48, 0xFFD438E4, 0xFF2948BD, 0xFF8F8EB2, 0xFFCCBE2F, 0xFF6DBAAE, 0xFF427469, 0xFFEA96E7, 0xFFD8C088, 0xFF067013, 0xFF5A7271, 0xFF29A815, 0xFFFDC4C5, 0xFF49A9F8, 0xFF8A098B, 0xFFEF7E70, 0xFF99FA3C, 0xFF073B3C, 0xFFB30812, 0xFF281F19, 0xFF7D8882, 0xFF13B8B6, 0xFF43AE33, 0xFFF9B22C, 0xFF056882, 0xFF6DEE79, 0xFFCE0F7E, 0xFF96BDF2, 0xFF0BAEC6, 0xFFD42A95, 0xFFA07A03, 0xFF436A10, 0xFF8EBD04, 0xFFD2EC4C, 0xFFC7E97C, 0xFF9A6FAE, 0xFF185A8D, 0xFF185732, 0xFFCF5AFE, 0xFFAD65D1, 0xFFF272F7, 0xFF4AF3AE, 0xFFAA0BF6, 0xFFF44957, 0xFFE8B3D7, 0xFF910CE2, 0xFF555F72, 0xFF5D6764, 0xFF952AF0, 0xFFD439E1, 0xFF132A7E, 0xFFC79E97, 0xFF22CA86, 0xFF556801, 0xFF45CF8A, 0xFF7313CA, 0xFFD97223, 0xFF76A0DA, 0xFF0DA4AD, 0xFF3EEC0D, 0xFFCAEBC5, 0xFF46912E, 0xFFFCA862, 0xFF2A9D2B, 0xFFD60B58, 0xFF6B8CFB, 0xFF022A66, 0xFF3B2D49, 0xFFF380CD, 0xFF021E00, 0xFF3F7739, 0xFF24BC08, 0xFF84684C, 0xFFC0122B, 0xFFB0F784, 0xFFA7372C, 0xFF47A7D3, 0xFF51BEB8, 0xFFAE04A8, 0xFF7A9F68, 0xFFFBC5A7, 0xFF549DB9, 0xFF1EEA18, 0xFFEB65FE, 0xFF014A01, 0xFF229605, 0xFF21F612, 0xFF4220DF, 0xFF67D4DF, 0xFF5F91BC, 0xFF849077, 0xFFB1386B, 0xFF8914D2, 0xFF1C6917, 0xFF99A982, 0xFF97C66E, 0xFFBB727D, 0xFFCE8A64, 0xFF17187F, 0xFF888321, 0xFFED224C, 0xFF5E3E2B, 0xFF1B57F9, 0xFFB492CC, 0xFFCD24A0, 0xFFA19706, 0xFFB8EDEA, 0xFF4D1712, 0xFF104681, 0xFF7DD266 ];
		TestImage(Data, 9, 18, Image);
	}



	/*
	Test decoding an image with multiple IDAT chunks.
	*/
	unittest
	{
		ubyte[] Data = [ 137, 80, 78, 71, 13, 10, 26, 10, 0, 0, 0, 13, 73, 72, 68, 82, 0, 0, 0, 3, 0, 0, 0, 2, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7, 73, 68, 65, 84, 120, 156, 99, 96, 96, 100, 2, 0, 0, 0, 0, 0, 0, 0, 7, 73, 68, 65, 84, 34, 102, 0, 0, 34, 0, 10, 0, 0, 0, 0, 0, 0, 0, 0, 73, 69, 78, 68, 0, 0, 0, 0 ];
		uint[] Image = [ 0xff000000, 0xff010101, 0xff020202, 0xff010101, 0xff020202, 0xff030303 ];
		TestImage(Data, 3, 2, Image);
	}
}